/*                                             -*- c-file-style: "Ellemtel" -*-
 *
 * mlink.c - Linux 2.6.x driver for M-Link board.
 *
 * Copyright 2005-2008
 * Authors: Alexey Shutov, Ilja Slepnev
 *
 * $HeadURL$
 * $Id$
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mount.h>
#include <linux/pci.h>
#include <linux/kobject.h>
#include <linux/sysfs.h>

#include <linux/kernel.h>       /* printk() */
#include <linux/fs.h>           /* everything... */
#include <linux/errno.h>        /* error codes */
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/printk.h>

#ifndef MLINK_PEX
# define MLINK_PEX
#endif

#define BAR_0    0
#define BAR_2    2
#define BAR_3    3

#include "debug.h"
#include "mlink.h"
#include "mlink_ioctls.h"

#include "mlink_op.h"
#include "mlink_op_pex.h"
#include "main.h"
#include "version.h"
#include "mlink_dev.h"
#include "frame_types.h"
#include "pexml_ports.h"
#include "compat.h"

mlink_status_t decode_link_status(hw_error_bits_t hw_error_bits)
{
    mlink_status_t st;
    st.all = 0;
    st.bit.link_down  = hw_error_bits.bit.rx_link_down;
    st.bit.data_error = hw_error_bits.bit.rx_data_error;
    st.bit.pci_error  = 0;
    st.bit.internal_error = hw_error_bits.bit.internal_error;
    st.bit.loss_of_signal = hw_error_bits.bit.loss_of_signal;
    st.bit.laser_fault = hw_error_bits.bit.laser_fault;
    st.bit.lcf_timeout = hw_error_bits.bit.lcf_timeout;
    return st;
}

typedef union {
  unsigned int all;
  struct {
      unsigned short minor : 16;
      unsigned short major : 16;
  } bit;
} version_t;

long dev_ioctl_wrapped (struct mlink_dev_t *dv, unsigned int cmd, unsigned long arg)
{
    //  printk("ioctl %08X, %s %s %ld bytes\n", cmd, (_IOC_DIR (cmd) & _IOC_READ)?"READ":"", (_IOC_DIR (cmd) & _IOC_WRITE)?"WRITE":"", size);
    switch (cmd) {

    case ML_IOC_RESET:
        dv_reset (dv);
        break;

    case ML_IOC_GET_DRIVER_VERSION:
        {
            version_t ver;
            ver.bit.major = VERSION_MAJOR_NUM;
            ver.bit.minor = VERSION_MINOR_NUM;
            put_user (ver.all, (unsigned long *) arg);
            break;
        }

    case ML_IOC_GET_SERIAL_ID:
        {
            uint64_t v = dv->board_info.board_serial;
            put_user (v, (uint64_t *) arg);
            break;
        }

//    case ML_IOC_GET_LATENCY:
//        {
//            put_user (dv->latency, (unsigned char *) arg);
//            break;
//        }

//    case ML_IOC_SET_LATENCY:
//        {
//            unsigned char c;

//            get_user (c, (unsigned char *) arg);
//            c &= ~0x07;
//            pci_write_config_byte (dv->pci_dev_brigde, PCI_LATENCY_TIMER, c);
//            pci_write_config_byte (dv->pci_dev_brigde, PCI_MIN_GNT, 8);
//            pci_write_config_byte (dv->pci_dev_brigde, PCI_MAX_LAT, 56);
//            dv->latency = (int) c;
//            PDEBUG("%s: PCI latency set to %d\n", dv->name, dv->latency);
//            break;
//        }

    case ML_IOC_GET_STATUS:
        {
            update_link_status(dv);
            mlink_status_t st = decode_link_status(dv->hw_error_bits);
            put_user (st, (mlink_status_t *) arg);
            break;
        }

    case ML_IOC_GET_CAPS:
        {
            mlink_caps_t st;
            st.all = 0;
            st.bit.serdes_loop_test = 1;
            st.bit.tx_cost = 2/2;
            st.bit.rx_cost = 1;
            put_user (st, (mlink_caps_t *) arg);
            break;
        }

    case ML_IOC_WRITE_PORT:
    // write 32-bit port register at given offset
        {
            ml_port_op_t op;
            int ret;

            if ((copy_from_user (&op, (void *) arg, sizeof (ml_port_op_t))))
                return (-EFAULT);
            ret = dv_write_port (dv, op.port, op.offset, op.data);
            if (ret)
                return (ret);
            break;
        }

    case ML_IOC_READ_PORT:
    // read 32-bit port register at given offset
        {
            ml_port_op_t op;
            int ret;
            // get userspace buffer address and size
            if ((copy_from_user (&op, (void *) arg, sizeof (ml_port_op_t))))
                return (-EFAULT);
            op.data = 0;
            ret = dv_read_port (dv, op.port, op.offset, &op.data);
            if (ret)
                return (ret);
//            put_user (op, (ml_port_op_t *) arg);
            if ((copy_to_user ((void *) arg, &op, sizeof (ml_port_op_t))))
                return (-EFAULT);
            break;
        }


    case ML_IOC_WRITE_FLASHER:
        {
            ml_flasher_op_t op;
            int ret;

            if ((copy_from_user (&op, (void *) arg, sizeof (ml_flasher_op_t))))
                return (-EFAULT);
            ret = dv_write_flasher (dv, op.offset, op.data);
            if (ret)
                return (ret);
            break;
        }

    case ML_IOC_READ_FLASHER:
        {
            ml_flasher_op_t op;
            int ret;
            if ((copy_from_user (&op, (void *) arg, sizeof (ml_flasher_op_t))))
                return (-EFAULT);
            op.data = 0;
            ret = dv_read_flasher (dv, op.offset, &op.data);
            if (ret)
                return (ret);
            if ((copy_to_user ((void *) arg, &op, sizeof (ml_flasher_op_t))))
                return (-EFAULT);
            break;
        }
    case ML_IOC_DMA_SEND:
        {
            ml_dmaop_ro_t dmaop;
            int ret;

            if ((copy_from_user (&dmaop, (void *) arg, sizeof (ml_dmaop_ro_t))))
                return (-EFAULT);
            ret = dv_send_dma (dv, dmaop);
            if (ret)
                return (ret);
            break;
        }

    case ML_IOC_DMA_RECV:
        {
            ml_dmaop_t dmaop;
            int ret;
            // get userspace buffer address and size
            if ((copy_from_user (&dmaop, (void *) arg, sizeof (ml_dmaop_t))))
                return (-EFAULT);
            ret = dv_recv_dma (dv, &dmaop);
            if (ret)
                return (ret);
            // update buffer info
            //    printk("ML_IOC_DMA_RECV: dmaop.size=%ld\n", dmaop.size);
            put_user (dmaop.size, (unsigned long *) &((ml_dmaop_t *) arg)->size);
            //      if ((copy_to_user (&dmaop, (void *) arg, sizeof (ml_dmaop_t)))) return(-EFAULT);
            break;
        }

    case ML_IOC_GET_EVENT:
        {
            ml_poll_ev_op_t op;
            int ret;
            // get parameters
            if ((copy_from_user (&op, (void *) arg, sizeof (ml_poll_ev_op_t))))
                return (-EFAULT);
            ret = dv_poll_event (dv, &op);
            if (ret)
                return (ret);
            // update info
            put_user (op.mask, (unsigned int *) &((ml_poll_ev_op_t *) arg)->mask);
            break;
        }

    case ML_IOC_READ_BOARD_INFO:
        {
            update_link_status(dv);
            update_board_id(dv);
            update_board_info(dv);
            if ((copy_to_user ((void *) arg, &dv->board_info, sizeof (board_info_t))))
                return (-EFAULT);
            break;
        }

    default:
        printk (KERN_ERR "%s: invalid ioctl %d\n", dv->name, cmd);
        return -EOPNOTSUPP;
    }
    return 0;
}

long dev_ioctl (struct file *file, unsigned int cmd, unsigned long arg)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
    struct inode *inode = file->f_dentry->d_inode;
#else
    struct inode *inode = file->f_path.dentry->d_inode;
#endif
    int minor = iminor(inode) & 0x0f;
    long size = _IOC_SIZE (cmd);
    struct mlink_dev_t *dv;
    long ret;

    if (minor >= device_count)
        return -ENXIO;
    dv = &dvtbl[minor];
    if (!dv->pid)
        return -ENXIO;              /* Shouldn't happen */
    if (!dv->initialized)
        return -EIO;                /* Hardware I/O error */

    /* Make sure we are passed a valid address pointer for arg. */
    if (arg != 0) {
        if (_IOC_DIR (cmd) & _IOC_READ) {
            int err = pexml_access_ok (VERIFY_WRITE, (void *) arg, size);
            if (err < 0) {
                PDEBUG ("ioctl VERIFY_WRITE: err=%d\n", err);
                return err;
            }
        }
        if (_IOC_DIR (cmd) & _IOC_WRITE) {
            int err = pexml_access_ok (VERIFY_READ, (void *) arg, size);
            if (err < 0) {
                PDEBUG ("ioctl VERIFY_READ: err=%d\n", err);
                return err;
            }
        }
    }

    mutex_lock(&dv->dev_mutex);
    ret = dev_ioctl_wrapped(dv, cmd, arg);
    mutex_unlock(&dv->dev_mutex);
    return ret;
}

/* ---------------------------------------------------------------
 * --- Map registers and request an IRQ (called once per card) ---
 * ---------------------------------------------------------------
 */
int dv_initialize (dev_p dv)
{
    int err = -1;
    unsigned long phys,           /* Physical start of the registers (not page aligned) */
    offset;                     /* phys - offset = page boundary for the registers */
    unsigned long phys2;           /* BAR_2 */
    size_t size;
    size_t i;
    int alloc_ok = 0;

    if (!dv || !dv->pci_dev_9056) {
        printk (KERN_ERR "dv struct not initialized\n");
        return -EIO;
    }

    if ((err = pci_enable_device(dv->pci_dev_9056))) {
        printk (KERN_ERR "%s: Failed to enable pci device (error %d)\n", dv->name, err);
        return (err);
    }

    /*
   * --- Memory-map resources and claim IO port regions ---
   */
//    phys = pci_resource_start (dv->pci_dev_brigde, 0);

//    if ((phys & PCI_BASE_ADDRESS_SPACE)) {
//        dv->err_status = "Initialization failed; not a memory mapped device";
//        printk (KERN_WARNING ALL_MSG "Not a memory mapped device.\n");
//        return -EIO;
//    }

    // Reserve DMA buffers
    alloc_ok = 1;
    dv->rxbuf.capacity = DV_FIFO_ENTRIES;
    for (i=0; i<dv->rxbuf.capacity; i++) {
        dv_fifo_ent_t *rxent = &dv->rxbuf.ent[i];
        rxent->size = DV_ENT_BYTES;
        rxent->buf = dma_alloc_coherent (&dv->pci_dev_9056->dev, rxent->size, &rxent->bus_addr, GFP_DMA);
        if (!rxent->buf) alloc_ok = 0;
    }

    if (!alloc_ok) {
        printk (KERN_WARNING "%s: Can't allocate RX DMA buffers\n", dv->name);
        for (i=0; i<dv->rxbuf.capacity; i++) {
            dv_fifo_ent_t *rxent = &dv->rxbuf.ent[i];
            if (rxent->buf)
                dma_free_coherent (&dv->pci_dev_9056->dev, rxent->size, rxent->buf, rxent->bus_addr);
            rxent->buf = NULL;
            rxent->size = 0;
            rxent->bus_addr = 0;
        }
        return (-ENOMEM);
    }

    dv->txbuf.size = 4096;        // *** FIXME ***
    dv->txbuf.virt_addr = dma_alloc_coherent (&dv->pci_dev_9056->dev, dv->txbuf.size, &dv->txbuf.bus_addr, GFP_DMA);

    if (!dv->txbuf.virt_addr) {
        printk (KERN_WARNING "%s: Can't allocate TX DMA buffer\n", dv->name);
        return (-ENOMEM);
    }

    for (i=0; i<dv->rxbuf.capacity; i++) {
        dv_fifo_ent_t *rxent = &dv->rxbuf.ent[i];
        memset(rxent->buf, 0, rxent->size);
    }
    memset(dv->txbuf.virt_addr, 0, dv->txbuf.size);

//    // Map I/O space for pci-express registers
//    phys &= PCI_BASE_ADDRESS_MEM_MASK;
//    offset = phys & ~PAGE_MASK;
//    size = PAGE_ALIGN (phys + sizeof (pex8311_peregs)) - (phys & PAGE_MASK);

//    dv->mreg_page = ioremap (phys & PAGE_MASK, size);
//    if (!dv->mreg_page) {
//        dv->err_status = "Initialization failed; couldn't ioremap registers";
//        printk (KERN_WARNING ALL_MSG "ioremap failed.\n");
//        return -EIO;
//    }
//    dv->mreg_pe = (pex8311_peregs *) ((caddr_t) dv->mreg_page + offset);


    phys = pci_resource_start (dv->pci_dev_9056, 0);
    if ((phys & PCI_BASE_ADDRESS_SPACE)) {
        dv->err_status = "Initialization failed; not a memory mapped device";
        printk (KERN_ERR "%s: Not a memory mapped device\n", dv->name);
        return -EIO;
    }

    // Map I/O space for local registers
    phys &= PCI_BASE_ADDRESS_MEM_MASK;
    offset = phys & ~PAGE_MASK;
    size = PAGE_ALIGN (phys + sizeof (pex8311_locregs)) - (phys & PAGE_MASK);

    dv->mreg_page_9056 = ioremap (phys & PAGE_MASK, size);
    if (!dv->mreg_page_9056) {
        dv->err_status = "Initialization failed; couldn't ioremap registers";
        printk (KERN_ERR "%s: ioremap failed\n", dv->name);
        return -EIO;
    }
    dv->mreg = (pex8311_locregs *) ((caddr_t) dv->mreg_page_9056 + offset);

    phys2 = pci_resource_start (dv->pci_dev_9056, BAR_2);
    dv->mem2 = ioremap (phys2 & PAGE_MASK , HW_ADDR_AREA_SIZE);
    if (!dv->mem2) {
        dv->err_status = "Initialization failed; couldn't ioremap mem2";
        printk (KERN_ERR "%s: mem2 ioremap failed\n", dv->name);
        return -EIO;
    }

    dv->mreg->lcs_las0ba = 0x1; //enable direct slave and DMA operations on local bus
    {
        u32 magic = dv_read_pexml_hw_reg(dv, HW_CSR_SYNC_WORD_REG);
        if (magic != 0xFEEDBEEF) {
            dv->err_status = "Initialization failed; invalid magic";
            printk (KERN_ERR "%s: Initialization failed; invalid magic %08X\n", dv->name, magic);
            return -EIO;
        }
    }
    update_board_id(dv);
    update_board_info(dv);
    print_board_info(dv);

    dv_setup_pci (dv);            /* Set bus mastering, latency, etc. */

    /*
   * --- Set up the Linux interrupt handler ---
   */

    dv->irq = dv->pci_dev_9056->irq;
    err = request_irq (dv->irq, irq_handler, IRQF_SHARED, dv->name, dv);

    if (err) {
        dv->err_status = "Initialization failed; couldn't get irq";
        printk (KERN_ERR "%s: Can't get irq %d\n", dv->name, dv->irq);
        return -EIO;
    }

    dv->hw_error_bits.all = 0;
    dv->save_bit_pci_error = 0;
    dv->rx_request_size_bytes = 0;
    dv->initialized = 1;
    dv->err_status = "OK";
//    dv->mreg_pe->gpio.all=0;

    dv->mreg->manycontrols.bit.loc_reset=1;
    dv->mreg->manycontrols.bit.loc_reset=0;

    dv_reset (dv);

    printk (KERN_DEBUG "%s: interrupt handler on irq %d\n", dv->name, dv->irq);
    return 0;
}

/* -------------------------------------------------------
 * --- Reset the AMCC S593X _and_ the M-Link interface ---
 * -------------------------------------------------------
 */

void dv_reset (dev_p dv)
{
    unsigned long flags;

    if (!dv || !dv->initialized)
        return;                     /* Shouldn't happen */

    spin_lock_irqsave (&dv->lock, flags);
    dv_xmit_disable (dv);
    dv->err_status = "Reset";

    rxbuf_clear(dv);
    txbuf_clear(dv);

    //reset hardware ram pointers
    dv->tx_wr_dataram_ptr = 0;
    dv->tx_wr_ptrram_ptr = 0;
    dv->rx_rd_dataram_ptr = 0;
    dv->rx_rd_ptrram_ptr = 0;

    printk (KERN_DEBUG "%s: Hardware reset\n", dv->name);
    dv->mreg->manycontrols.bit.loc_reset=1;  // Assert LRESET# - FPGA reset
    dv_clear_status(dv);
    dv->mreg->manycontrols.bit.loc_reset=0;  // Deassert LRESET#

    // DMA0: TX, DMA1: RX

    dv->mreg->dmamode0.bit.locam=0; // Set increasing local address for dma0
    dv->mreg->dmamode0.bit.rie=1;   // Enable READY# input for dma0
    dv->mreg->dmamode1.bit.locam=0;  // Set increasing local address for dma1
    dv->mreg->dmamode1.bit.rie=1;    // Enabe READY# input for dma1
    dv->mreg->dmamode0.bit.lbe=1;    // enable local burst mode for DMA 0
    dv->mreg->dmamode1.bit.lbe=1;    // enable local burst mode for DMA 1
    dv->mreg->dmamode0.bit.contbe=1; // enable cont. local burst mode for DMA 0
    dv->mreg->dmamode1.bit.contbe=1; // enable cont. local burst mode for DMA 1
    dv->mreg->dmadp0 = 0x0;   // set only direction PE->LOC for DMA 0
    dv->mreg->dmadp1 = 0x8;   // set only direction LOC->PE for DMA 1
    dv->mreg->dmala0 = 0x40000;   // random local address
    dv->mreg->dmala1 = 0x40000;   // random local address
    dv->mreg->dmamode0.bit.intsel = 1;    // PCIE interrupt for DMA 0 (0 - to LOCAL)
    dv->mreg->dmamode1.bit.intsel = 1;    // PCIE interrupt for DMA 1 (0 - to LOCAL)
    dv->mreg->dmamode0.bit.doneie = 1;   // int. on dma done, set when dma is programmed
    dv->mreg->dmamode1.bit.doneie = 1;   // int. on dma done, set when dma is programmed
    dv->mreg->dmamode1.bit.demm = 0;   // dis. demand mode from original driver

    dv->mreg->lintcsr.bit.dma0ie=1;   // enable interrupt on dma 0 complete
    dv->mreg->lintcsr.bit.dma1ie=1;   // enable interrupt on dma 1 complete
    dv->mreg->lintcsr.bit.liie=1;     // enable local interrupt for incoming mailbox interrupt
    dv->mreg->lintcsr.bit.mbie=1;     // enable local interrupt for outgoing mailbox

    dv->mreg->dmaarb=0x00C0000;     // enable BREQi for bus arbitration
    dv->mreg->lcs_las0ba = 0x1; //enable direct slave and DMA operations on local bus

    dv->max_rx_frame_bytes = ML_MAX_RAW_FRAME_WORDS*4;
    dv->max_tx_frame_bytes = ML_MAX_RAW_FRAME_WORDS*4;

    // Enable LCF interrupts
    dv_set_irq_mask_bits(dv, LOCAL_IRQ_SRC_MASK_LCF_BIT, 0, FORCE_UPDATE);
    dv_write_pexml_hw_reg(dv, 0xFFFFFFFF, HW_RX_LCF_ACK_DATA_REG);  //unreal lcf_ack_data so we don't miss first lcf

    update_board_id(dv);

    // reset ports
    if (dv_caps_gtx_ctrl(dv)) {
        unsigned int port;
        for (port=1; port<=4; port++) {
            const unsigned int gtx_enable  = HW_CSR_PORT_CONTROL_BITS_GTX_ENABLE;
            const unsigned int gtx_disable = HW_CSR_PORT_CONTROL_BITS_GTX_DISABLE;
            const unsigned int sfp_enable  = (1 << SFP_CTRL_BIT_RT_SEL);
            const unsigned int sfp_disable = (1 << SFP_CTRL_BIT_TX_DISABLE);
            // disable port
            dv_write_port (dv, port, PEXML_GTX_CTRL_REG, gtx_disable);
            dv_write_port (dv, port, PEXML_SFP_CTRL_REG, sfp_disable);
            // enable ports 1, 4
            if (port == 1) {
                dv_write_port (dv, port, PEXML_GTX_CTRL_REG, gtx_enable);
                dv_write_port (dv, port, PEXML_GTX_CTRL_REG, gtx_enable | GTX_CTRL_RESET);
                dv_write_port (dv, port, PEXML_GTX_CTRL_REG, gtx_enable);
            } else {
                dv_write_port (dv, port, PEXML_GTX_CTRL_REG, gtx_disable);
            }
            dv_write_port (dv, port, PEXML_SFP_CTRL_REG, (port == 1) ? sfp_enable : sfp_disable);
        }
    }

    // Enable RX
    dv_try_start_rx(dv);
    spin_unlock_irqrestore (&dv->lock, flags);

    printk (KERN_DEBUG "%s: Hardware reset complete\n", dv->name);
}

void dv_unmap_mreg(dev_p dv)
{
    if (dv->mreg_page_9056) {
        iounmap (dv->mreg_page_9056);    /* Unmap the operation registers */
        dv->mreg_page_9056 = NULL;
        dv->mreg = NULL;
    }
    if (dv->mem2) {
        iounmap (dv->mem2);              /* Unmap BAR_2 memory  */
        dv->mem2 = NULL;
    }
}

void dv_interrupt_locked_handle_imb1 (dev_p dv)
{
    // run in locked interrupt handler
    dv_recv_imb (dv);

    if (dv->imb1.all == 0xFFFFFFFF) {
        dv->pci_error = 1;
        dv_int_disable(dv);
        dv_xmit_disable(dv); // FIXME
        dv->err_status = "Device is offline: IMB=FFFFFFFF";
        return;
    }

    // Handle LCF change
    {
        uint32_t lcf_reg = 0;
        lcf_reg = dv_read_pexml_hw_reg(dv, HW_RX_LCF_DATA_REG);
        if (dv->imb1.bit.lcf_data_ready) {
            lcf_reg = dv_read_pexml_hw_reg(dv, HW_RX_LCF_DATA_REG); //read lcf data
            dv->lcf_data.all = lcf_reg & 0xFF;
            dv_write_pexml_hw_reg(dv, lcf_reg, HW_RX_LCF_ACK_DATA_REG); //ack. lcf data
            dv->imb_event++;
            dv->wake_imb++;
            dv->n_irq_imb_lcf++;
        }
    }

    // incoming packet handling should be the very last action.
    if (dv->imb1.bit.rx_dma_data_ready) {
        dv->rx_event++;
        dv->n_irq_imb_rx++;
    }
    // Enable LCF interrupts
    dv_set_irq_mask_bits(dv, LOCAL_IRQ_SRC_MASK_LCF_BIT, 0, OPTIMAL_UPDATE);
}


static int dv_interrupt_locked_handle_rtc (dev_p dv)
{
    dv->rtc_event = 1;            //frame send
    dv_read_disable (dv);
    if (dv->txbuf.used > 0) {
        dv->txbuf.flags |= DMA_BUF_FLAG_COMPLETE;
        dv->stats.tx_packets++;
        dv->stats.tx_bytes += dv->txbuf.used;
        dv->wake_tx++;
    }
    return 0;
}

unsigned int dv_rxfifo_dirty_used(dv_rxfifo_t buf)
{
    // circular buffer size = (capacity + head - tail) % capacity
    return (unsigned int)(buf.capacity + buf.ndirty - buf.nwr) % buf.capacity;
}

static const int DUMP_ERROR_FRAME_HEADER = 1;
static const int DUMP_ERROR_FRAME_FULL = 0;

int validate_rxbuf_entry(dev_p dv, rxfifo_ptr t_ptr, uint16_t debug_index, uint16_t debug_total)
{
    int error_flag = 0;
    const unsigned char *buf = rxbuf_frame(&dv->rxbuf, t_ptr);
    unsigned int buf_bytes = rxbuf_frame_len(&dv->rxbuf, t_ptr);
    unsigned int p=0;
    unsigned int fn=0;
    if (buf_bytes < 4*ML_MIN_RAW_FRAME_WORDS) {
        printk (KERN_WARNING "%s: rx buffer %u/%u (index %2u, %5u bytes) has invalid size\n",
                dv->name, debug_index, debug_total, t_ptr, buf_bytes);
        error_flag = 1;
    }
    while (p + sizeof(frame_hdr_t) < buf_bytes) {
        const unsigned char *offset = p+buf;
        const frame_hdr_t *fh = (const frame_hdr_t *)offset;
        uint16_t expected_seq = (dv->save_frame_seq + 1u) & 0xFFFF;
        if (0) printk (KERN_DEBUG "%s: rx buffer %u/%u (index %2u, %5u bytes), offset %04X: frame %u sync=%04X type=%04X seq=%d len=%d\n",
                       dv->name, debug_index, debug_total, t_ptr, buf_bytes, p,
                       fn, fh->sync, fh->type, fh->seq, fh->len);
        if (fh->sync != ML_FRAME_SYNC) {
            printk (KERN_WARNING "%s: rx buffer %u/%u (index %2u, %5u bytes), offset %04X: frame %u sync: %04X\n",
                    dv->name, debug_index, debug_total, t_ptr, buf_bytes, p,
                    fn, fh->sync);
            if (DUMP_ERROR_FRAME_HEADER)
                print_hex_dump(KERN_DEBUG, ALL_MSG  " frame header: ", DUMP_PREFIX_OFFSET, 32, 4, fh, sizeof(frame_hdr_t), 1);
            if (DUMP_ERROR_FRAME_FULL)
                print_hex_dump(KERN_DEBUG, ALL_MSG  " rx buffer: ", DUMP_PREFIX_OFFSET, 32, 4, buf, buf_bytes, 1);
            if (fh->sync == ML_FRAME_SYNC) {
                printk (KERN_WARNING "%s: rx buffer %u/%u (index %2u, %5u bytes), offset %04X: frame %u sync: %04X\n",
                        dv->name, debug_index, debug_total, t_ptr, buf_bytes, p,
                        fn, fh->sync);
            }
            error_flag = 1;
            break;
        }
        if (fh->len < ML_MIN_RAW_FRAME_WORDS || fh->len > ML_MAX_RAW_FRAME_WORDS) {
            printk (KERN_WARNING "%s: rx buffer %u/%u (index %2u, %5u bytes), offset %04X: bad frame %u len: %u\n",
                    dv->name, debug_index, debug_total, t_ptr, buf_bytes, p,
                    fn, fh->len);
            error_flag = 1;
            break;
        }
        if (!error_flag && fh->type == FRAME_TYPE_VME_DATA) {
            // check for sequence number
            if (dv->save_frame_seq_init && fh->seq != expected_seq) {
                uint16_t missed = fh->seq - expected_seq;
                if (missed < 256)
                    printk (KERN_NOTICE "%s: rx buffer %u/%u (index %2u, %5u bytes), offset %04X: missed %d frames: previous seq %u, current %u\n",
                            dv->name, debug_index, debug_total, t_ptr, buf_bytes, p,
                            missed, dv->save_frame_seq, fh->seq);
                else
                    printk (KERN_NOTICE "%s: rx buffer %u/%u (index %2u, %5u bytes), offset %04X: bad sequence number in received frame type %04X: %u, expected %u\n",
                            dv->name, debug_index, debug_total, t_ptr, buf_bytes, p,
                            fh->type, fh->seq, expected_seq);
                dv->stats.rx_missed_errors++;
            }
            dv->save_frame_seq = fh->seq;
            dv->save_frame_seq_init = 1;
        }
        p += fh->len * 4;
        fn++;
    }
    if (error_flag) {
        dv->save_frame_seq_init = 0;
    }
    return error_flag;
}

// multiple buffers may be received in one interrupt
void validate_rxbuf(dev_p dv, rxfifo_ptr tail, rxfifo_ptr head)
{
    // circular buffer size = (capacity + head - tail) % capacity
    uint16_t rxbuf_received_entries = (dv->rxbuf.capacity + head - dv->rxbuf.nwr) % dv->rxbuf.capacity;
    rxfifo_ptr t_ptr = tail;
    uint16_t bn=1;
    while (t_ptr != head) {
        validate_rxbuf_entry(dv, t_ptr, bn, rxbuf_received_entries);
        t_ptr = (t_ptr + 1) % dv->rxbuf.capacity;
        bn++;
    }
}

// RX
static int dv_interrupt_locked_handle_wtc (dev_p dv)
{
    int error_flag = 0;
    rxfifo_ptr next_nwr = 0;

#ifdef DEBUG
    // expect enabled transfer
    if (!dv->mreg->dmacs1.bit.dmace) {
        printk(KERN_ERR "%s: wtc: dmace=0", dv->name);
        error_flag = 1;
        return error_flag;
    }
#endif
    // write-transfer-complete interrupt serialization
    dv->mreg->dmacs1.bit.dmace = 0;

    PDEBUG("%s: wtc: RX DMA complete for %5u bytes: nwr=%2u, ndirty=%2u\n",
           dv->name, dv->mreg->dmats1, dv->rxbuf.nwr, dv->rxbuf.ndirty);

    next_nwr = dv->rxbuf.ndirty;
    /*
    // multiple buffers may be received in one interrupt
    {
        unsigned int rxbuf_received_entries = dv_rxfifo_dirty_used(dv->rxbuf);
        int n = 1;
        rxfifo_ptr t_ptr = dv->rxbuf.nwr;
        while (t_ptr != next_nwr) {
            unsigned int data_size = rxbuf_frame_len(&dv->rxbuf, t_ptr);
            if (1) PDEBUG("wtc: received buffer %u/%u (index %2u), len %u\n",
                   n, rxbuf_received_entries, t_ptr, data_size);
            t_ptr = (t_ptr + 1) % dv->rxbuf.capacity;
            n++;
        }
    }
    */
    // sanity check - first frame in first buffer must be valid
    {
        // pointer to header of first frame in first received buffer
        const frame_hdr_t *fh0 = rxbuf_writer_frame_header(&dv->rxbuf);
        if (fh0->sync != ML_FRAME_SYNC) {
            dv->stats.rx_errors++;
            printk (KERN_WARNING "%s: RX DMA buffer contains invalid data: first frame in buffer has sync %04X, expected %04X\n",
                    dv->name, fh0->sync, ML_FRAME_SYNC);
            error_flag = 1;
        }
    }
    // reset sequence number validator
    if (error_flag) {
        dv->save_frame_seq_init = 0;
    }
    // validate headers of received frames
    if (!error_flag) {
        validate_rxbuf_entry(dv, dv->rxbuf.nwr, 1, 1);
    }
    //write pointers reg to HW
    dv_write_pexml_hw_reg(dv, dv->rx_rd_dataram_ptr, HW_RX_RD_DATARAM_PTR_REG);
    dv_write_pexml_hw_reg(dv, dv->rx_rd_ptrram_ptr, HW_RX_RD_PTRRAM_PTR_REG);

    // advance nwr
    dv->rxbuf.nwr = next_nwr;
    dv->wtc_event = 1;
    dv->wake_rx++;
    dv_try_start_rx(dv); // LOOK AT ME

    return error_flag;
}


/// dv_interrupt_locked() is run inside spinlock.
/// Do not use printk, user-code access, any functions that may sleep or you will DEADLOCK
int dv_interrupt_locked (dev_p dv)
{
    lintcsr reg_lintcsr;
    register int int_source_tx = 0;  //0-dma0
    register int int_source_rx = 0;  //1-dma1
    register int int_source_imb = 0;  //2-incoming mailbox
    register int int_source_other = 0;

//    test_pin_up(dv);

    dv->irq_count++;
    reg_lintcsr.all = dv->mreg->lintcsr.all;

    // Power Management Interrupt
    if (reg_lintcsr.bit.pmi) {
        reg_lintcsr.bit.pmi = 1;
        int_source_other = 1;
        dv->n_irq_other++;
    }

    // local bus parity error
    //  if(dv->mreg_loc->lintcsr.bit.dmwrpces) {
    //    dv->mreg_loc->lintcsr.bit.dmwrpces=1;
    //  }

    // PCI doorbell interrupt
    if(reg_lintcsr.bit.pcidia) {
        int_source_other = 1;
        dv->n_irq_other++;
    }

    // PCI abort interrupt
    if(reg_lintcsr.bit.pciaia) {
        int_source_other = 1;
        dv->n_irq_other++;
    }

    // TX complete interrupt
    if (reg_lintcsr.bit.dma0ia)
    {
        dv->mreg->dmacs0.bit.dmaci = 1;
        int_source_tx = 1;
        dv->n_irq_tx++;
    }

    // RX complete interrupt
    if (reg_lintcsr.bit.dma1ia)
    {
        dv->mreg->dmacs1.bit.dmaci = 1;
        int_source_rx = 1;
        dv->n_irq_rx++;
    }

    // Mailbox interrupt
    if (reg_lintcsr.bit.lia)
    {
        ml_clear_imb_interrupt_request(dv);
        int_source_imb = 1;
        dv->n_irq_imb++;
    }

    // Unknown interrupt
    if (! (int_source_tx||int_source_rx||int_source_imb||int_source_other)) {
//        test_pin_down(dv);
        return 0;
    }

    if (int_source_imb) {
        dv_interrupt_locked_handle_imb1 (dv);
    }
    if (dv->mreg->dmacs0.bit.dmadone == 1 && int_source_tx) {
        dv_interrupt_locked_handle_rtc (dv);
    }
    if (dv->mreg->dmacs1.bit.dmadone == 1 && int_source_rx) {
        dv_interrupt_locked_handle_wtc (dv);
    }

//    test_pin_down(dv);
    return 1;
}

/// dv_interrupt_bottom_half is not protected by spinlock.
/// It may use functions that sleep: printk, kalloc.
void dv_interrupt_bottom_half (dev_p dv)
{
    report_errors (dv);
    dv_interrupt_check_pci_errors (dv);
    if (dv->rx_event) { dv_receive_packets(dv); dv->rx_event--;}
    if (dv->imb_event) { update_link_status(dv); dv->imb_event--;}

    if (dv->wake_rx > 1) {
        PDEBUG("dv->wake_rx = %d\n", dv->wake_rx);
    }
    if (dv->wake_tx > 1) {
        PDEBUG("dv->wake_tx = %d\n", dv->wake_tx);
    }
    if (dv->wake_imb > 1) {
        PDEBUG("dv->wake_imb = %d\n", dv->wake_imb);
    }
    while (dv->wake_rx > 0) {
        wake_up_interruptible (&dv->waitq);
        dv->wake_rx--;
    }
    while (dv->wake_tx > 0) {
        wake_up_interruptible (&dv->waitq);
        dv->wake_tx--;
    }
    while (dv->wake_imb > 0) {
        wake_up_interruptible (&dv->waitq);
        dv->wake_imb--;
    }
}

/* End of main_pex.c */
