//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "main.h"
#include "debug.h"
#include "mlink_dev.h"
#include "mlink_op.h"
#include "mlink_op_pex.h"
#include "version.h"
#include "version_str.h"
#include "ml_proc.h"
#include "pexml_ports.h"

#define ML_PROCFS_NAME ML_DEVICE_NAME

//const char *pexml_stat_str(int n) {
//    switch (n) {
//        case 0:  return "Link transition";
//        case 1:  return "RX packet";
//        case 2:  return "RX CRC error";
//        case 3:  return "RX code error";
//        case 4:  return "RX OSD";
//        case 5:  return "RX alignment error";
//        case 6:  return "RX data error";
//        case 7:  return "RX LCF timeout";
//        case 8:  return "RX OSD";
//        case 9:  return "TX packet";
//        case 10: return "TX CRC error";
//        case 11: return "RX code error";
//        case 12: return "[12]";
//        case 13: return "[13]";
//        case 14: return "[14]";
//        case 15: return "[15]";
//        default: return "unnamed";
//    }
//}

static void mlink_proc_show_speed(struct seq_file *m, dev_p dv)
{
    const u32 irq_period_us = dv_read_pexml_hw_reg(dv, HW_CSR_IRQ_MIN_PERIOD_US_REG);
    const u32 clk_khz  = dv_read_pexml_hw_reg(dv, HW_CSR_CLK_KHZ_REG);
    const u32 clk_mult = dv_read_pexml_hw_reg(dv, HW_CSR_CLK_MULT_REG);
    const u32 clk_div  = dv_read_pexml_hw_reg(dv, HW_CSR_CLK_DIV_REG);
    const u32 data_bits = 16;
    const u32 sysclk_khz = 62500;
    const u32 link_speed = sysclk_khz * data_bits * clk_mult / clk_div / 1000;
    seq_printf(m, "  Link speed: %d Mbps\n", link_speed);
}

static void mlink_proc_show_gtx_sfp(struct seq_file *m, dev_p dv)
{
    unsigned int i;
    const char *gtx_state_str[8] = {
        "INIT",
        "CPLL_RESET",
        "WAIT_CPLL_LOCK",
        "WAIT_RDY_TX",
        "WAIT_RDY_RX",
        "RX_RESET",
        "RUN",
        "UNKNOWN-7"
    };
    for (i=0; i<4; i++) {
        unsigned int port = i+1;
        u32 gtx_ctrl = 0, gtx_stat = 0, sfp_ctrl = 0, sfp_stat = 0;
        dv_read_port (dv, port, PEXML_GTX_CTRL_REG, &gtx_ctrl);
        dv_read_port (dv, port, PEXML_GTX_STAT_REG, &gtx_stat);
        dv_read_port (dv, port, PEXML_SFP_CTRL_REG, &sfp_ctrl);
        dv_read_port (dv, port, PEXML_SFP_STAT_REG, &sfp_stat);

        u32 gtx_rxpd = (gtx_ctrl & GTX_CTRL_RXPD) != 0;
        u32 gtx_txpd = (gtx_ctrl & GTX_CTRL_TXPD) != 0;
        u32 gtx_loopback = (gtx_ctrl & GTX_CTRL_LOOPBACK) != 0;
        u32 gtx_pd = gtx_rxpd || gtx_txpd;
        u32 trx_present = (sfp_stat & ( 1 << SFP_STAT_BIT_MOD_DETECT)) == 0;
        u32 los         = (sfp_stat & ( 1 << SFP_STAT_BIT_LOS));
        u32 tx_fault    = (sfp_stat & ( 1 << SFP_STAT_BIT_TX_FAULT));
        u32 tx_rt_sel   = (sfp_ctrl & ( 1 << SFP_CTRL_BIT_RT_SEL));
        u32 tx_disable  = (sfp_ctrl & ( 1 << SFP_CTRL_BIT_TX_DISABLE));
        seq_printf(m, "  port[%d]: GTX %04X/%04X %s%s%s", port,
                   gtx_ctrl, gtx_stat,
                   gtx_rxpd ? " RXPD": "",
                   gtx_txpd ? " TXPD": "",
                   gtx_loopback ? " LOOPBACK" : "");
        if (!gtx_pd)
            seq_printf(m, " %s (%s%s%s%s%s)",
                       gtx_state_str[(gtx_stat & 0x07)],
                    (gtx_stat & 0x08) ? " tx_reset_done" : "",
                    (gtx_stat & 0x10) ? " rx_reset_done" : "",
                    (gtx_stat & 0x20) ? " cpll_locked" : "",
                    (gtx_stat & 0x40) ? " gtx_ready" : "",
                    (gtx_stat & 0x80) ? " rx_link_up" : ""
                                  );
        seq_printf(m, ", SFP: %02X/%02X %s%s%s%s%s\n",
                   sfp_ctrl, sfp_stat,
                   trx_present ? "trx_present" : "trx_absent",
                   (trx_present && tx_rt_sel) ? " rt_sel" : "",
                   (trx_present && tx_disable) ? " tx_disable" : "",
                   (trx_present && los) ? " signal_lost" : "",
                   (trx_present && tx_fault)? " tx_fault" : ""
                                              );
    }
    return;
}

static void mlink_proc_show_rx_counters(struct seq_file *m, dev_p dv)
{
    // statistics
    u32 cnt_rx_link_transitions = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 0);
    u32 cnt_rx_packet           = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 1);
    u32 cnt_rx_crc_error        = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 2);
    u32 cnt_rx_code_error       = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 3);
    u32 cnt_rx_osd_transitions  = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 4);
    u32 cnt_rx_align_error      = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 5);
    u32 cnt_rx_data_error       = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 6);
    u32 cnt_rx_lcf_timeout      = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + 7);
    seq_printf(m, "  RX hardware link transitions: %d\n", cnt_rx_link_transitions);
    seq_printf(m, "  RX hardware packets: %d\n", cnt_rx_packet);
    seq_printf(m, "  RX hardware CRC errors: %d\n", cnt_rx_crc_error);
    seq_printf(m, "  RX hardware alignment errors: %d\n", cnt_rx_align_error);
    seq_printf(m, "  RX hardware code errors: %d\n", cnt_rx_code_error);
    seq_printf(m, "  RX hardware data errors: %d\n", cnt_rx_data_error);
    seq_printf(m, "  RX hardware LCF timeouts: %d\n", cnt_rx_lcf_timeout);
    seq_printf(m, "  RX hardware OSD transitions: %d\n", cnt_rx_osd_transitions);
}

static int mlink_proc_show(struct seq_file *m, void *v) {
    int i;

    seq_printf (m, "driver version : %s\n", VERSION_STR);
    seq_printf (m, "device count : %d\n", device_count);
#if 0
    seq_printf(m, "major number   : %d\n", device_major);
#endif
    for (i = 0; i < device_count; i++) {
        dev_p dv;

        dv = &dvtbl[i];
        if (!dv)
            continue;
        update_link_status(dv);
        seq_printf(m, "/dev/%.32s\n", dv->name);
        seq_printf(m, "  Board info         : serial %04X-%04X-%04X, firmware %d.%d.%d\n",
                   (u16)((dv->board_info.board_serial >> 32) & 0xFFFF),
                   (u16)((dv->board_info.board_serial >> 16) & 0xFFFF),
                   (u16)(dv->board_info.board_serial & 0xFFFF),
                   dv->board_info.fw_version.fw_ver_maj,
                   dv->board_info.fw_version.fw_ver_min,
                   dv->board_info.fw_version.fw_rev
                   );
        seq_printf(m, "  Buffers            : RX %d bytes / %d packets, TX %d bytes / %d packets\n",
                   4 * dv->rx_dataram_size, dv->rx_ptrram_size, 4 * dv->tx_dataram_size, dv->tx_ptrram_size);

        seq_printf(m, "  Interrupt %-3d      : %lu, IMB:%lu (LCF:%lu, RX:%lu), DMA_TX:%lu, DMA_RX:%lu, other:%lu\n", dv->irq,
                   dv->irq_count, dv->n_irq_imb, dv->n_irq_imb_lcf, dv->n_irq_imb_rx, dv->n_irq_tx, dv->n_irq_rx, dv->n_irq_other);
        seq_printf(m, "  Received LCF data  : %02X %s%s%s%s%s%s%s%s\n",
            dv->lcf_data.all,
            dv->lcf_data.bit.remote_loopback ? "LOOPBACK "        : "",
            dv->lcf_data.bit.remote_error    ? "REMOTE_ERROR "    : "",
            dv->lcf_data.bit.remote_ready    ? "REMOTE_READY "    : "",
            dv->lcf_data.bit.remote_rx_error ? "REMOTE_RX_ERROR " : "",
            dv->lcf_data.bit.spill           ? "SPILL "           : "",
            dv->lcf_data.bit.suspend         ? "XOFF "            : "",
            dv->lcf_data.bit.remote_rx_crc_error ? "Remote_CRC_Error " : "",
            dv->lcf_data.bit.remote_link_up      ? "Remote_Link_Up "  : ""
            );

        seq_printf(m, "  Error bits         : %02X %s%s%s%s%s%s%s%s%s\n",
            dv->hw_error_bits.all,
            dv->hw_error_bits.all ? "" : "Ok",
            dv->hw_error_bits.bit.internal_error ? "Internal_Error " : "",
            dv->hw_error_bits.bit.rx_link_down   ? "Link_Down "      : "",
            dv->hw_error_bits.bit.rx_data_error  ? "RX_Data_Error "  : "",
            dv->hw_error_bits.bit.loss_of_signal ? "Signal_Loss "    : "",
            dv->hw_error_bits.bit.laser_fault    ? "Laser_Fault "    : "",
            dv->hw_error_bits.bit.lcf_timeout    ? "LCF_Timeout "    : "",
            dv->hw_error_bits.bit.error_bit_6    ? "error_6 "        : "",
            dv->hw_error_bits.bit.error_bit_7    ? "error_7 "        : "");

        seq_printf(m, "  Last error         : %s\n", dv->err_status);
        {
            s16 ow_temp = dv_read_pexml_hw_reg(dv, HW_CSR_TEMP_REG) & 0xFFF;
            if (ow_temp != 0xFFF) {
                s16 temp, temp_int, temp_dec;
                if (ow_temp & 0x800) ow_temp = -(ow_temp & 0x7FF);
                temp = ow_temp * 10 / 16;
                temp_int = temp / 10;
                temp_dec = temp - temp_int * 10;
                seq_printf(m, "  Board temperature  : %d.%01d C\n",
                temp_int, temp_dec);
            }
        }
        seq_printf(m, "\n");
        seq_printf(m, "  RX  b:%-8lu pkt:%-8lu err:%-8lu drop:%-8lu over:%-8lu frame:%-8lu\n",
                        dv->stats.rx_bytes, dv->stats.rx_packets, dv->stats.rx_errors, dv->stats.rx_dropped, dv->stats.rx_over_errors, dv->stats.rx_frame_errors);
        seq_printf(m, "  TX  b:%-8lu pkt:%-8lu err:%-8lu drop:%-8lu over:%-8lu carr.:%-8lu\n",
                        dv->stats.tx_bytes, dv->stats.tx_packets, dv->stats.tx_errors, dv->stats.tx_dropped, dv->stats.tx_fifo_errors, dv->stats.tx_carrier_errors);
        seq_printf(m, "\n");

        seq_printf(m, "  TX DMA buffer: %9s, size %ld, used %ld, count %ld, seq %ld, flags %08lX %s%s\n",
                       txbuf_allocated(dv)?"allocated":"null", dv->txbuf.size, dv->txbuf.used, dv->txbuf.count, dv->txbuf.seq,
                       dv->txbuf.flags,
                       (dv->txbuf.flags & DMA_BUF_FLAG_BUSY)?"BUSY ":"",
                       (dv->txbuf.flags & DMA_BUF_FLAG_COMPLETE)?"DMA_COMPLETE ":"");

        seq_printf(m, "  RX DMA buffer: %u x %u kB, ndma-nwr-nrd: %u-%u-%u, avail %u, unread %u, flags %08lX %s%s\n",
                       dv->rxbuf.capacity, dv->rxbuf.capacity?dv->rxbuf.ent[0].size/1024:0,
                       //  	  dv->rxbuf.used, dv->rxbuf.count,
                       //  	  dv->rxbuf.used,
                       dv->rxbuf.ndirty, dv->rxbuf.nwr, dv->rxbuf.nrd,
                       rxbuf_dirty_avail(dv),
                       rxbuf_unread_count(dv),
                       //  	  dv->rxbuf.seq,
                       dv->rxbuf.flags,
                       (dv->rxbuf.flags & DMA_BUF_FLAG_BUSY)?"BUSY ":"",
                       (dv->rxbuf.flags & DMA_BUF_FLAG_COMPLETE)?"DMA_COMPLETE ":"");

        seq_printf(m, "\n");

#ifdef DEBUG
        if (dv->mreg) {
            ssp_imb1_t imb;
            imb.all = dv->mreg->mbox1;
            /*
            seq_printf(m, "  imb=%08X  RxCnt=%04X  ErrorBits=%02X  LCF=%02X\n",
                imb.all,
                imb.bit.rxcnt,
                imb.bit.hw_error_bits.all,
                imb.bit.lcf_data.all);

                */
            seq_printf(m, "  lintcsr: %s%s\n",
                           dv->mreg->lintcsr.bit.dma0ie?"dma0ie ":"",
                           dv->mreg->lintcsr.bit.dma1ie?"dma1ie ":""
                           );
            // 0: TX,   1: RX
            seq_printf(m, "  TX dma: %s ts=%u pci=%08X loc=%08X %s\n",
                           dv->mreg->dmamode0.bit.doneie?"done0ie ":"",
                           dv->mreg->dmats0,
                           dv->mreg->dmapcia0,
                           dv->mreg->dmala0,
                           dv->mreg->dmacs0.bit.dmadone ? "done" : ""
                       );
            seq_printf(m, "  RX dma: %s ts=%u pci=%08X loc=%08X %s\n",
                           dv->mreg->dmamode1.bit.doneie?"done1ie ":"",
                           dv->mreg->dmats0,
                           dv->mreg->dmapcia1,
                           dv->mreg->dmala1,
                           dv->mreg->dmacs1.bit.dmadone ? "done" : ""
                           );
        }
#endif
        if (dv_caps_gtx_ctrl(dv)) {
            mlink_proc_show_speed(m, dv);
            mlink_proc_show_gtx_sfp(m, dv);
        }
//        if (0) {
//            int j;
//            for (j=0; j<16; j++) {
//                u32 cnt = dv_read_pexml_hw_reg(dv, HW_STAT_REGION_ADDR + j);
//                if (cnt != 0)
//                    seq_printf(m, "%-16s: %u\n", pexml_stat_str(j), cnt);
//            }
//        }
        mlink_proc_show_rx_counters(m, dv);

#ifdef DEBUG
        {
            uint32_t rx_wr_ptrram_ptr = dv_read_pexml_hw_reg(dv, HW_RX_WR_PTRRAM_PTR_REG); //read number of writen frames
            uint32_t pkt_count = (dv->rx_rd_ptrram_ptr == rx_wr_ptrram_ptr) ? 0 : (rx_wr_ptrram_ptr - 1 - dv->rx_rd_ptrram_ptr) % dv->rx_ptrram_size;
            seq_printf(m, "  RX pkt_ram: rd_ptr %d, wr_ptr %d, count %d, req_bytes %d\n",
                       dv->rx_rd_ptrram_ptr, rx_wr_ptrram_ptr, pkt_count, dv->rx_request_size_bytes);
        }
        if (0) {
            uint32_t tx_wr_ptrram_ptr = dv_read_pexml_hw_reg(dv, HW_TX_WR_PTRRAM_PTR_REG);
            uint32_t tx_wr_dataram_ptr = dv_read_pexml_hw_reg(dv, HW_TX_WR_DATARAM_PTR_REG);
            seq_printf(m, "  TX pkt_ram: wr_ptr %d\n", tx_wr_ptrram_ptr);
            seq_printf(m, "  TX data_ram: wr_ptr %d\n", tx_wr_dataram_ptr);
        }
#endif
    }
    return 0;
}

static int mlink_proc_open(struct inode *inode, struct  file *file) {
  return single_open(file, mlink_proc_show, NULL);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
static const struct file_operations mlink_proc_ops = {
  .owner = THIS_MODULE,
  .open = mlink_proc_open,
  .read = seq_read,
  .llseek = seq_lseek,
  .release = single_release,
};
#else
static const struct proc_ops mlink_proc_ops = {
  .proc_open = mlink_proc_open,
  .proc_read = seq_read,
  .proc_lseek = seq_lseek,
  .proc_release = single_release,
};
#endif

void mlink_proc_init(void) {
  proc_create(ML_PROCFS_NAME, 0, NULL, &mlink_proc_ops);
}

void mlink_proc_exit(void) {
  remove_proc_entry(ML_PROCFS_NAME, NULL);
}
