//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MLINK_DRIVER_MLINK_DEV_H
#define MLINK_DRIVER_MLINK_DEV_H


#ifndef MLINK_PEX
# define MLINK_PEX
#endif

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#include <linux/mutex.h>
#include <linux/cdev.h>

#include "PEX8311.h"
#include "mlink.h"
#include "mlink_types.h"

struct mlink_dev_t {
    struct pci_dev *pci_dev_9056; /* The result of pci_get_device() */
    struct cdev cdev;             /* Char device structure      */
    int id;                       /* 0 or 1 or ... */
    char name[32];                /* Name of the entry in "/dev/" */
    int irq;                      /* Interrupt level for this card */
    unsigned long irq_count;      /* Number of interrupts since initialization */
    int latency;                  /* PCI latency timer for this card */
    int initialized;              /* Only set to 1 if initialize completes */
    pid_t pid;             /* 0=closed, pid=original PID that open()'ed */
    caddr_t mreg_page_9056;            /* Saved only for iounmap() */
    spinlock_t lock;              /* Used by the interrupt handler */
    uint32_t* mem2;
    /*
   * These are "volatile" because interrupt handlers might change them
   * or, in the case of the registers, they talk to memory-mapped IO.
   */
    volatile pex8311_locregs *mreg;    /* Memory mapped PLX 9056 regs */
    volatile const char *err_status;  /* String with last error or "OK" */
    volatile unsigned int dma_errors; /* ++ if anything bad happens */
    volatile unsigned int dma_report; /* compare with dma_errors outside of RT */

    wait_queue_head_t waitq;      /* Used by bottom-half handler */
    int wake_tx, wake_rx, wake_imb; /* Set by rt handler to signal bh */

    volatile ssp_imb1_t imb1;
    volatile uint32_t irq_mask;

    uint32_t rx_request_size_bytes;
    volatile int rtc_event;
    volatile int wtc_event;
    volatile int imb_event;
    volatile int rx_event;

    hw_error_bits_t hw_error_bits;
    int pci_error; // there was pci error since reset
    int save_bit_pci_error;
    dv_rxfifo_t rxbuf; /* RX buffers */
    dma_buf_t txbuf;   /* small buffer for TX DMA */ // FIXME: TX FIFO
    net_device_stats_t stats;       /* network device statistics */
    // irq counts
    unsigned long n_irq_imb, n_irq_imb_lcf, n_irq_imb_rx, n_irq_tx, n_irq_rx, n_irq_other;
    unsigned long max_rx_frame_bytes;
    unsigned long max_tx_frame_bytes;
    lcf_data_t lcf_data, save_lcf_data;
    // frame validation data
    uint16_t save_frame_seq_init;
    uint16_t save_frame_seq;
    struct mutex dev_mutex;
    //hardware RAM pointers
    uint32_t tx_wr_dataram_ptr;
    uint32_t tx_wr_ptrram_ptr;
    uint32_t rx_rd_dataram_ptr;
    uint32_t rx_rd_ptrram_ptr;
    uint32_t tx_dataram_size;
    uint32_t tx_ptrram_size;
    uint32_t rx_dataram_size;
    uint32_t rx_ptrram_size;
    // board info
    board_info_t board_info;
};

typedef struct mlink_dev_t *dev_p;

enum { HW_ADDR_AREA_SIZE = 1024*1024 }; //1Mb predefined, see manual
//enum { TX_DATARAM_SIZE = 1024 }; //hardware defined
//enum { TX_PTRRAM_SIZE = 8 }; //hardware defined
//enum { RX_DATARAM_SIZE = 8192 }; //hardware defined
//enum { RX_PTRRAM_SIZE = 64 }; //hardware defined

// Note 1: hardware limit on mapped region size 1 MByte
// mmap region is divided by 4 x 256 kByte regions: registers, rx_core, tx_core, dummy
// Note 2: register hardware address * 4 = memory address
//
enum {
    HW_CSR_REGION_ADDR   = 0x0,
    HW_STAT_REGION_ADDR  = 0x200,
    HW_FLASHER_REGION_ADDR = 0x7000,
    HW_RX_REGION_ADDR    = 0x10000, // 256-512 kBytes
    HW_TX_REGION_ADDR    = 0x20000  // 512-768 kBytes
};

enum { HW_CSR_REGION_MASK     = 0x000001FFu};
enum { HW_STAT_REGION_MASK    = 0x000001FFu};
enum { HW_FLASHER_REGION_MASK = 0x000000FFu};
enum { HW_RX_REGION_MASK      = 0x0000FFFFu};
enum { HW_TX_REGION_MASK      = 0x0000FFFFu};

enum {
    HW_DATA_RAM_ADDR_OFFSET              = 0x0,
    HW_PTR_RAM_ADDR_OFFSET               = 0x4000, // 0x8000,
    HW_RAM_POINTERS_REGS_ADDR_OFFSET     = 0x8000  // 0xC000
};

enum {
    HW_WR_DATARAM_PTR_REG             = 0x0,
    HW_WR_PTRRAM_PTR_REG              = 0x1,
    HW_RD_DATARAM_PTR_REG             = 0x2,
    HW_RD_PTRRAM_PTR_REG              = 0x3
};

enum {
    HW_TX_DATA_RAM_BASE_ADDR    = HW_TX_REGION_ADDR | HW_DATA_RAM_ADDR_OFFSET,
    HW_TX_PTR_RAM_BASE_ADDR     = HW_TX_REGION_ADDR | HW_PTR_RAM_ADDR_OFFSET,
    HW_TX_WR_DATARAM_PTR_REG    = HW_TX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_WR_DATARAM_PTR_REG,
    HW_TX_WR_PTRRAM_PTR_REG     = HW_TX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_WR_PTRRAM_PTR_REG,
    HW_TX_RD_PTRRAM_PTR_REG     = HW_TX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_RD_PTRRAM_PTR_REG
};

enum {
    HW_RX_DATA_RAM_BASE_ADDR    = HW_RX_REGION_ADDR | HW_DATA_RAM_ADDR_OFFSET,
    HW_RX_PTR_RAM_BASE_ADDR     = HW_RX_REGION_ADDR | HW_PTR_RAM_ADDR_OFFSET,
    HW_RX_WR_DATARAM_PTR_REG    = HW_RX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_WR_DATARAM_PTR_REG,
    HW_RX_WR_PTRRAM_PTR_REG     = HW_RX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_WR_PTRRAM_PTR_REG,
    HW_RX_RD_DATARAM_PTR_REG    = HW_RX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_RD_DATARAM_PTR_REG,
    HW_RX_RD_PTRRAM_PTR_REG     = HW_RX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | HW_RD_PTRRAM_PTR_REG
};

enum {
    HW_CSR_PEX_CORE_STATUS_REG        = HW_CSR_REGION_ADDR | 0x00,
    HW_CSR_SYNC_WORD_REG              = HW_CSR_REGION_ADDR | 0x01,
    HW_CSR_IRQ_SRC_REG                = HW_CSR_REGION_ADDR | 0x02,
    HW_CSR_IRQ_SRC_EN_MASK_REG        = HW_CSR_REGION_ADDR | 0x03,
    HW_CSR_RX_STATUS                  = HW_CSR_REGION_ADDR | 0x04,
    HW_CSR_TX_STATUS                  = HW_CSR_REGION_ADDR | 0x05,
    HW_CSR_BOARD_CONFIG_REG           = HW_CSR_REGION_ADDR | 0x06,
    HW_CSR_IRQ_MIN_PERIOD_US_REG      = HW_CSR_REGION_ADDR | 0x07,
    HW_CSR_CLK_KHZ_REG                = HW_CSR_REGION_ADDR | 0x08,
    HW_CSR_CLK_MULT_REG               = HW_CSR_REGION_ADDR | 0x09,
    HW_CSR_CLK_DIV_REG                = HW_CSR_REGION_ADDR | 0x0A,
    HW_CSR_OW_ID_LO_REG               = HW_CSR_REGION_ADDR | 0x50,
    HW_CSR_OW_ID_HI_REG               = HW_CSR_REGION_ADDR | 0x51,
    HW_CSR_TEMP_REG                   = HW_CSR_REGION_ADDR | 0x54,
    HW_CSR_FW_VER_REG                 = HW_CSR_REGION_ADDR | 0x56,
    HW_CSR_FW_REV_REG                 = HW_CSR_REGION_ADDR | 0x57
};

enum {
    HW_RX_LCF_DATA_REG = HW_RX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | 0x4,
    HW_RX_LCF_ACK_DATA_REG = HW_RX_REGION_ADDR | HW_RAM_POINTERS_REGS_ADDR_OFFSET | 0x5
};

enum {
    LOCAL_IRQ_SRC_MASK_RX_DMA_BIT   = 0x1,
    LOCAL_IRQ_SRC_MASK_LCF_BIT      = 0x2
};

enum {
    OPTIMAL_UPDATE = 0,
    FORCE_UPDATE = 1
};

#endif // MLINK_DRIVER_MLINK_DEV_H
