//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/module.h>

#include <linux/uaccess.h>
#include <linux/sched.h>

#include "debug.h"
#include "mlink_op.h"
#include "mlink_op_pex.h"
#include "mlink_frame_header.h"
#include "compat.h"

int txbuf_allocated(const dev_p dv)
{
    return (dv->txbuf.virt_addr && dv->txbuf.size>0);
}
// ..buf_has_space(): return 1 if it can hold 1 or more frames
int txbuf_has_space(const dev_p dv, unsigned long nbytes)
{
    if (!txbuf_allocated(dv)) return 0;
    return (dv->txbuf.used + nbytes <= dv->txbuf.size);
}
// number of packets available for dma
unsigned int rxbuf_dirty_avail(const dev_p dv)
{
    // read pointers once to local variables
    const rxfifo_ptr nrd = dv->rxbuf.nrd; // may be modified in userspace context
    const rxfifo_ptr ndirty = dv->rxbuf.ndirty;

    rxfifo_ptr avail;
    if (nrd == ndirty)
        avail = dv->rxbuf.capacity;
    else
        avail = (rxfifo_ptr)(nrd - ndirty) % dv->rxbuf.capacity;
    return avail;
}

unsigned int rxbuf_unread_count(const dev_p dv)
{
    rxfifo_ptr count;
    count = (rxfifo_ptr)(dv->rxbuf.nwr - dv->rxbuf.nrd) % dv->rxbuf.capacity;
    return count;
}

int txbuf_has_data(const dev_p dv)
{
    if (!txbuf_allocated(dv)) return 0;
    return (dv->txbuf.count > 0);
}

void txbuf_clear(dev_p dv)
{
//    unsigned long flags;
//    spin_lock_irqsave (&dv->lock, flags);
    dv->txbuf.nrd = 0;
    dv->txbuf.nwr = 0;
    dv->txbuf.flags = 0;
    dv->txbuf.used = 0;
    dv->txbuf.count = 0;
//    spin_unlock_irqrestore (&dv->lock, flags);
}

void rxbuf_clear(dev_p dv)
{
//    unsigned long flags;
//    spin_lock_irqsave (&dv->lock, flags);
    dv->rxbuf.ndirty = 0;
    dv->rxbuf.nrd = 0;
    dv->rxbuf.nwr = 0;
    dv->rxbuf.flags = 0;
//    spin_unlock_irqrestore (&dv->lock, flags);
}

//------------------------------------------------------------------------------

// must be run inside spinlock!
unsigned int dv_try_start_rx(dev_p dv)
{
    unsigned int has_space = rxbuf_dirty_avail(dv);
    //enable rx data ready local interrupt(all other are already enabled in dv_interrupt_locked_handle_imb1, so we use EN_ALL),
    // else(no buf available) keep it disabled
    if (has_space > 1)
        dv_set_irq_mask_bits(dv, LOCAL_IRQ_SRC_MASK_RX_DMA_BIT, 0, OPTIMAL_UPDATE);
    return has_space;
}

// LOCKED: check if space is available and enable next packet rx
// return positive if ack has been sent
int dv_locked_try_start_rx(dev_p dv)
{
    unsigned long flags;
    unsigned int has_space = 0;
    spin_lock_irqsave (&dv->lock, flags);
    has_space = dv_try_start_rx(dv);
    spin_unlock_irqrestore (&dv->lock, flags);
    if (!has_space) {
        printk(KERN_ERR "%s: dv_locked_try_start_rx: no space for next packet\n", dv->name);
    }
    return has_space;
}

int dv_recv_dma (dev_p dv, ml_dmaop_t * dmaop)
{
    static const char *LOCAL_PREFIX = "IOC_DMA_RECV:";

    int err = pexml_access_ok (VERIFY_READ, dmaop, sizeof(void*));
    if (err < 0) {
        return (err);
    }

//    if (dmaop->timeout >= 0) {
        // wait for incoming frame

#define rx_event_condition (!(dv->rxbuf.nwr == dv->rxbuf.nrd))
        long wait_ret = 0;
        int timeout_ticks = (dmaop->timeout * HZ)/1000;
        if (timeout_ticks == 0)
            timeout_ticks = 1;
//            wait_ret = wait_event_interruptible (dv->waitq, rx_event_condition);
//        else
            wait_ret = wait_event_interruptible_timeout (dv->waitq, rx_event_condition, timeout_ticks);

        if (timeout_ticks>0 && wait_ret == 0) {
            //                PDEBUG("%s RX timeout\n", LOCAL_PREFIX);
            return (-ETIME);
        }

        if (wait_ret == -ERESTARTSYS) {
            PDEBUG("ERESTARTSYS\n");
            return (-ERESTARTSYS);
        }
        //      if (signal_pending (current)) return -ERESTARTSYS;

//    } else {
//        if (!rx_event_condition) {
//            PDEBUG("No available packets *************************\n");
//            // try receive next frame if rx fifo has space
//            dv_locked_try_start_rx(dv);
//            return -EAGAIN;
//        }
//    }// end of if (timeout >= 0)


    const unsigned long received_frame_bytes = rxbuf_frame_len(&dv->rxbuf, dv->rxbuf.nrd);

    if (dmaop->size < received_frame_bytes) {
        printk (KERN_ERR "%s not enough space: packet size is %ld bytes, buffer is %ld bytes, truncated\n", LOCAL_PREFIX, received_frame_bytes, dmaop->size);
    } else {
        PDEBUGG("received_frame_bytes %lu from entry %lu\n", received_frame_bytes, dv->rxbuf.nrd);
        dmaop->size = received_frame_bytes;
    }

    err = pexml_access_ok (VERIFY_WRITE, (void *) dmaop->buf, dmaop->size);
    if (err < 0) {
        // drop packets
        printk (KERN_ERR "%s access_ok=%d\n", LOCAL_PREFIX, err);
        goto cleanup;
    }
    // move packet to userspace
    err = copy_to_user (dmaop->buf, rxbuf_frame(&dv->rxbuf, dv->rxbuf.nrd), dmaop->size);
    if (err) {
        printk (KERN_ERR "%s copy_to_user=%d\n", LOCAL_PREFIX, err);
        goto cleanup;
    }

cleanup:
    dv->rxbuf.nrd = (dv->rxbuf.nrd + 1) % dv->rxbuf.capacity;
    //    PDEBUG("dv->rxbuf.nrd++, now %lu\n", dv->rxbuf.nrd);

    // try receive next frame ONLY if fifo is empty
    if (!rx_event_condition) {
        dv_locked_try_start_rx(dv);
    }
    return 0;
}

const unsigned char *rxbuf_frame(const dv_rxfifo_t *rxbuf, rxfifo_ptr index) {
    return rxbuf->ent[index % rxbuf->capacity].buf;
}

unsigned int rxbuf_frame_len(const dv_rxfifo_t *rxbuf, rxfifo_ptr index) {
    return rxbuf->ent[index % rxbuf->capacity].len;
}

// allocated buffer size at index
unsigned int rxbuf_buf_size(const dv_rxfifo_t *rxbuf, rxfifo_ptr index) {
    return rxbuf->ent[index % rxbuf->capacity].size;
}

const frame_hdr_t *rxbuf_writer_frame_header(const dv_rxfifo_t *rxbuf) {
    return (const frame_hdr_t *) rxbuf_frame(rxbuf, rxbuf->nwr);
}
