//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MLINK_DRIVER_MLINK_TYPES_H
#define MLINK_DRIVER_MLINK_TYPES_H

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/types.h>
#include <asm/atomic.h>

//
// For this time, we use char device, not network device.
// Some code taken from <linux/netdevice.h> follows:

/*
 *	Network device statistics. Akin to the 2.0 ether stats but
 *	with byte counters.
 */

typedef struct
{
    unsigned long	rx_packets;		/* total packets received	*/
    unsigned long	tx_packets;		/* total packets transmitted	*/
    unsigned long	rx_bytes;		/* total bytes received 	*/
    unsigned long	tx_bytes;		/* total bytes transmitted	*/
    unsigned long	rx_errors;		/* bad packets received		*/
    unsigned long	tx_errors;		/* packet transmit problems	*/
    unsigned long	rx_dropped;		/* no space in linux buffers	*/
    unsigned long	tx_dropped;		/* no space available in linux	*/
    unsigned long	multicast;		/* multicast packets received	*/
    unsigned long	collisions;

    /* detailed rx_errors: */
    unsigned long	rx_length_errors;
    unsigned long	rx_over_errors;		/* receiver ring buff overflow	*/
    unsigned long	rx_crc_errors;		/* recved pkt with crc error	*/
    unsigned long	rx_frame_errors;	/* recv'd frame alignment error */
    unsigned long	rx_fifo_errors;		/* recv'r fifo overrun		*/
    unsigned long	rx_missed_errors;	/* receiver missed packet	*/

    /* detailed tx_errors */
    unsigned long	tx_aborted_errors;
    unsigned long	tx_carrier_errors;
    unsigned long	tx_fifo_errors;
    unsigned long	tx_heartbeat_errors;
    unsigned long	tx_window_errors;

    /* for cslip etc */
    unsigned long	rx_compressed;
    unsigned long	tx_compressed;
} net_device_stats_t;

enum {
    DMA_BUF_FLAG_COMPLETE = 0x00000002,
    DMA_BUF_FLAG_BUSY = 0x00000004,
};

typedef struct {
    unsigned long size;
    unsigned long used; // number of bytes busy
    unsigned long count; // number of bytes received/transmitted
    unsigned long seq; // sequence number
    unsigned long flags;
    unsigned long nrd; // number of writes
    unsigned long nwr; // number of reads
    dma_addr_t bus_addr;
    void *virt_addr;
} dma_buf_t;

// rx/tx packet FIFO
//enum { DV_FIFO_ENTRIES = 16 }; // size must be a divider of 256
enum { DV_FIFO_ENTRIES = 64 }; // size must be a divider of 256
enum { DV_ENT_BYTES = ML_RX_BUFFER_SIZE };

typedef struct {
    //  unsigned long offset; // bytes offset of raw packet start in dmabuf
    unsigned int size; // allocated size, bytes
    unsigned int len; // packet length, bytes
    unsigned char *buf;
    dma_addr_t bus_addr; // bus address of buf
} dv_fifo_ent_t;

typedef volatile unsigned char rxfifo_ptr;

typedef struct {
    rxfifo_ptr capacity; // total number of entries == DV_FIFO_ENTRIES
    rxfifo_ptr ndirty; // DMA writer entry index (hardware)
    rxfifo_ptr nwr; // writer entry index (hardware)
    rxfifo_ptr nrd; // reader entry index (user)
    dv_fifo_ent_t ent[DV_FIFO_ENTRIES];
    unsigned long flags;
} dv_rxfifo_t;

typedef volatile unsigned int int_count_t;

// hardware errors
typedef union {
    uint8_t all;
    struct {
        unsigned int internal_error: 1; // LSERR# | !dcm_lock | laser_fault | rx_fifo_overflow
        unsigned int rx_link_down:  1;  // !rx_clock | !link_up
        unsigned int rx_data_error: 1;  // bad_sync | bad_length | alignment | CRC | 8b10b_violation
        unsigned int loss_of_signal: 1; // LOS#
        unsigned int laser_fault: 1;    // O_FAULT
        unsigned int lcf_timeout: 1;    // lcf_timeout
        unsigned int error_bit_6: 1;    // reserved error bit 6
        unsigned int error_bit_7: 1;    // reserved error bit 7
    } __attribute__ ((packed)) bit;
} hw_error_bits_t;

// LCF status
typedef union
{
    uint8_t all;
    struct {
        unsigned int suspend: 1;
        unsigned int remote_loopback: 1;
        unsigned int spill: 1;
        unsigned int remote_error: 1;
        unsigned int remote_ready: 1;
        unsigned int remote_rx_error: 1;
        unsigned int remote_rx_crc_error: 1;
        unsigned int remote_link_up: 1;
    } __attribute__ ((packed))  bit;
} lcf_data_t;

// Incoming Mailbox Data
typedef union
{
    uint32_t all;
    struct
    {
        unsigned int rx_dma_data_ready : 1; // rx data ready to transmit from board to driver
        unsigned int lcf_data_ready    : 1; // LCF data ready to update
        unsigned int hw_error_bits     : 1; // look for error bits in corr. CSR_REG(currently not implemented)
    } __attribute__ ((packed)) bit;
} ssp_imb1_t;

#endif // MLINK_DRIVER_MLINK_TYPES_H
